'
' The replacement for the static System.IO.Path class that works with
' files/directories that are greater than 260 characters in length
'
Public Class PathEx

#Region "API Region"
    'DWORD WINAPI GetFullPathName(
    '  __in          LPCTSTR lpFileName,
    '  __in          DWORD nBufferLength,
    '  __out         LPTSTR lpBuffer,
    '  __out         LPTSTR* lpFilePart
    ');
    Private Declare Auto Function GetFullPathName Lib "kernel32" ( _
        ByVal lpFileName As String, _
        ByVal nBufferLength As Integer, _
        ByVal lpBuffer As System.Text.StringBuilder, _
        ByVal lpFilePart As Integer _
    ) As Integer
#End Region

    Public Shared Function GetDirectoryName(ByVal Path As String) As String
        Dim ans, parts() As String

        If Path = "\" Then
            Return ""
        End If

        If Path.Length = 3 And Path.EndsWith(":\") Then
            ' this seems counter intuitive, but I don't make the rules
            Return ""
        End If

        parts = Path.Split("\"c)

        ' bare naked UNC path?
        If Path.StartsWith("\\") And parts.Length = 4 Then
            Return ""
        End If

        If parts.Length > 0 Then
            ans = String.Join("\", parts, 0, parts.Length - 1)
            If ans = "" And parts.Length > 1 Then
                Return "\"
            End If

            ' Note: can't end with a colon
            If ans.Length = 2 AndAlso ans.EndsWith(":") Then
                ans &= "\"
            End If
            Return ans
        End If

        Return ""
    End Function

    Public Shared Function GetExtension(ByVal Path As String) As String
        Dim parts() As String

        If Path.EndsWith(".") Then
            Return ""
        End If

        parts = GetFileName(Path).Split("."c)
        If parts.Length > 1 Then
            ' Note: includes the starting "."
            Return "." & parts(UBound(parts))
        End If

        Return ""
    End Function

    Public Shared Function GetFileName(ByVal Path As String) As String
        Dim parts() As String

        ' just a drive letter?
        If Path.Length = 2 And Path.EndsWith(":") Then
            Return ""
        End If

        parts = Path.Split("\"c)
        Return parts(UBound(parts))
    End Function

    Public Shared Function GetFileNameWithoutExtension(ByVal Path As String) As String
        Dim index As Integer
        Dim fname As String

        fname = GetFileName(Path)
        index = fname.LastIndexOf(".")
        If index >= 0 Then
            Return fname.Substring(0, index)
        End If

        Return fname
    End Function

    Public Shared Function GetFullPath(ByVal Path As String) As String
        Dim ret As Integer
        Dim ans As New System.Text.StringBuilder(260)

        ' Start with a buffer that's 260 in length
        ret = GetFullPathName(Path, 260, ans, 0)
        If ret > 260 Then
            ' resize the buffer, and try again
            ans.Capacity = ret
            GetFullPathName(Path, ret, ans, 0)
        End If

        ' Note: the *real* Path.GetFullPath() method throws an exception if
        ' there are illegal characters in the path.  That's not really 
        ' appropriate here, so I'm not following their lead.
        Return ans.ToString
    End Function

    Public Shared Function GetPathRoot(ByVal Path As String) As String
        Dim parts() As String

        If IsPathRooted(Path) = False Then
            Return ""
        End If

        ' just a drive letter and colon?
        If Path.Length = 2 And Path.EndsWith(":") Then
            Return Path
        End If

        parts = Path.Split("\"c)

        ' Is this a UNC Path?
        If Path.StartsWith("\\") And parts.Length >= 4 Then
            ' Note: doesn't end with a trailing backslash
            Return String.Join("\", parts, 0, 4)
        End If

        Return String.Join("\", parts, 0, 1) & "\"
    End Function

    Public Shared Function HasExtension(ByVal Path As String) As Boolean
        Return CBool(GetExtension(Path) <> "")
    End Function

    Public Shared Function IsPathRooted(ByVal Path As String) As Boolean

        ' does it start with a drive letter and colon?
        If Path.Length >= 2 AndAlso Path.Substring(1, 1) = ":" Then
            Return True
        End If

        ' absolute path (or UNC path)
        If Path.StartsWith("\") Then
            Return True
        End If

        Return False
    End Function
End Class
