Imports System.IO
Imports System.Security.AccessControl
Imports System.Security.Principal

Public Class Traverse
    Private fm As fmMain
    Private first, dir_only, consolidate, needs_conversion As Boolean
    Private dt As DataTable
    Private db As AccessDatabase
    Private server, RootName As String
    Private f_count, d_count, row_count As Long

    Private Enum AceFlags As Byte
        OBJECT_INHERIT_ACE = &H1
        CONTAINER_INHERIT_ACE = &H2
        NO_PROPAGATE_INHERIT_ACE = &H4
        INHERIT_ONLY_ACE = &H8
        INHERITED_ACE = &H10
        VALID_INHERIT_FLAGS = &H1F
        SUCCESSFUL_ACCESS_ACE_FLAG = &H40
        FAILED_ACCESS_ACE_FLAG = &H80
    End Enum

    Private Enum AccessTypes As Integer
        DELETE = &H10000
        READ_CONTROL = &H20000
        WRITE_DAC = &H40000
        WRITE_OWNER = &H80000
        SYNCHRONIZE = &H100000
        STANDARD_RIGHTS_REQUIRED = &HF0000
        STANDARD_RIGHTS_READ = READ_CONTROL
        STANDARD_RIGHTS_WRITE = READ_CONTROL
        STANDARD_RIGHTS_EXECUTE = READ_CONTROL
        STANDARD_RIGHTS_ALL = &H1F0000
        SPECIFIC_RIGHTS_ALL = &HFFFF
        ACCESS_SYSTEM_SECURITY = &H1000000
        MAXIMUM_ALLOWED = &H2000000
        GENERIC_READ = &H80000000
        GENERIC_WRITE = &H40000000
        GENERIC_EXECUTE = &H20000000
        GENERIC_ALL = &H10000000
        FILE_LIST_DIRECTORY = &H1
        FILE_READ_DATA = &H1
        FILE_WRITE_DATA = &H2
        FILE_ADD_FILE = &H2
        FILE_APPEND_DATA = &H4
        FILE_ADD_SUBDIRECTORY = &H4
        FILE_READ_EA = &H8
        FILE_WRITE_EA = &H10
        FILE_TRAVERSE = &H20
        FILE_EXECUTE = &H20
        FILE_DELETE_CHILD = &H40
        FILE_READ_ATTRIBUTES = &H80
        FILE_WRITE_ATTRIBUTES = &H100
        FILE_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED Or SYNCHRONIZE Or &H1FF
        FILE_GENERIC_READ = STANDARD_RIGHTS_READ Or FILE_LIST_DIRECTORY Or FILE_READ_ATTRIBUTES Or FILE_READ_EA Or SYNCHRONIZE
        FILE_GENERIC_WRITE = STANDARD_RIGHTS_WRITE Or FILE_ADD_FILE Or FILE_WRITE_ATTRIBUTES Or FILE_WRITE_EA Or FILE_ADD_SUBDIRECTORY Or SYNCHRONIZE
        FILE_GENERIC_EXECUTE = STANDARD_RIGHTS_EXECUTE Or FILE_READ_ATTRIBUTES Or FILE_TRAVERSE Or SYNCHRONIZE
    End Enum

    Public Property MainForm() As fmMain
        Get
            Return fm
        End Get
        Set(ByVal value As fmMain)
            fm = value
        End Set
    End Property

    Public Property IsFirst() As Boolean
        Get
            Return first
        End Get
        Set(ByVal value As Boolean)
            first = value
        End Set
    End Property

    Public Property ConsolidateDirs() As Boolean
        Get
            Return Consolidate
        End Get
        Set(ByVal value As Boolean)
            consolidate = value
        End Set
    End Property

    Public Property DirOnly() As Boolean
        Get
            Return dir_only
        End Get
        Set(ByVal value As Boolean)
            dir_only = value
        End Set
    End Property

    Public Property DataTable() As DataTable
        Get
            Return dt
        End Get
        Set(ByVal value As DataTable)
            dt = value
        End Set
    End Property

    Public Property AccessDatabase() As AccessDatabase
        Get
            Return db
        End Get
        Set(ByVal value As AccessDatabase)
            db = value
        End Set
    End Property

    Public Property ServerName() As String
        Get
            Return server
        End Get
        Set(ByVal value As String)
            server = value
        End Set
    End Property

    Public Property RecordRootName() As String
        Get
            Return RootName
        End Get
        Set(ByVal value As String)
            RootName = value
        End Set
    End Property

    Public ReadOnly Property FileCount() As Long
        Get
            Return f_count
        End Get
    End Property

    Public ReadOnly Property DirectoryCount() As Long
        Get
            Return d_count
        End Get
    End Property

    Public ReadOnly Property RowCount() As Long
        Get
            Return row_count
        End Get
    End Property

    '
    ' The main entry point for this class... it reads all of the files/directories
    ' and records the data into the database
    '
    Public Sub Doit(ByVal StartingDir As String)
        Dim p_di, di, di_array() As DirectoryInfoEx
        Dim ds As DirectorySecurityEx
        Dim fs As FileSecurityEx
        Dim fi, fi_array() As FileInfoEx
        Dim fsar As System.Collections.ObjectModel.ReadOnlyCollection(Of FileSystemAccessRule)
        Dim LastDACL As New List(Of FileSystemAccessRule)
        Dim isDifferent As Boolean
        Dim i As Integer
        Dim dr As DataRow

        ' dress it up a wee bit...
        If Not StartingDir.EndsWith("\") Then
            StartingDir = StartingDir & "\"
        End If

        ' skip the recycle bin and other directories we don't care about...
        For Each skip As String In My.Settings.SkipList
            If StartingDir.EndsWith(skip, StringComparison.CurrentCultureIgnoreCase) Then
                Exit Sub
            End If
        Next

        Try
            p_di = New DirectoryInfoEx(StartingDir)
        Catch
            ' silently ignore these errors
            Exit Sub
        End Try

        ' keep 'em entertained...
        fm.lbStatus.Text = StartingDir
        Application.DoEvents()

        ' get the starting directory info (first run only)
        Try
            If first Then
                ' do we need to convert a UNC path?
                needs_conversion = CBool(Not (StartingDir.StartsWith(RootName)))

                ds = p_di.GetAccessControl()
                LastDACL.Clear()
                For Each ar As FileSystemAccessRule In ds.GetAccessRules(True, True, GetType(SecurityIdentifier))
                    dr = dt.NewRow
                    dr("Server") = server
                    dr("Path") = FormatPath(StartingDir)
                    dr("Type") = ar.AccessControlType.ToString
                    dr("Name") = TranslateSID.TranslateSidToName(server, CType(ar.IdentityReference, SecurityIdentifier))
                    dr("Permissions") = DirectoryMaskToString(ar.FileSystemRights)
                    dr("Inherited") = ar.IsInherited
                    dr("Scope") = ACEFlagToString(ar.InheritanceFlags, ar.PropagationFlags, ar.IsInherited)
                    dt.Rows.Add(dr)
                    LastDACL.Add(ar)
                Next

                d_count += 1
                first = False
            Else
                If consolidate Then
                    ' save the DACL of the current directory
                    ds = p_di.GetAccessControl()
                    LastDACL.Clear()
                    For Each ar As FileSystemAccessRule In ds.GetAccessRules(True, True, GetType(SecurityIdentifier))
                        LastDACL.Add(ar)
                    Next
                End If
            End If
        Catch ex As Exception
            dr = dt.NewRow
            dr("Server") = server
            dr("Path") = FormatPath(StartingDir)
            dr("Name") = Left("Error: " & ex.Message, 128)
            dt.Rows.Add(dr)
            ' not fatal
        End Try


        If Not dir_only Then
            ' check for errors in files before we begin
            Try
                fi_array = p_di.GetFiles
            Catch ex As Exception
                dr = dt.NewRow
                dr("Server") = server
                dr("Path") = FormatPath(StartingDir)
                dr("Name") = Left("Error: " & ex.Message, 128)
                dt.Rows.Add(dr)
                Exit Sub
            End Try

            ' process files before we do directories
            For Each fi In fi_array
                Try
                    fs = fi.GetAccessControl()
                    For Each ar As FileSystemAccessRule In fs.GetAccessRules(True, True, GetType(SecurityIdentifier))
                        dr = dt.NewRow
                        dr("Server") = server
                        dr("Path") = FormatPath(fi.FullName)
                        dr("Type") = ar.AccessControlType.ToString
                        dr("Name") = TranslateSID.TranslateSidToName(server, CType(ar.IdentityReference, SecurityIdentifier))
                        dr("Permissions") = FileMaskToString(ar.FileSystemRights)
                        dr("Inherited") = ar.IsInherited
                        ' files don't use the Scope field
                        dt.Rows.Add(dr)
                    Next
                    f_count += 1
                Catch ex As Exception
                    dr = dt.NewRow
                    dr("Server") = server
                    dr("Path") = FormatPath(fi.FullName)
                    dr("Name") = Left("Error: " & ex.Message, 128)
                    dt.Rows.Add(dr)
                    ' not fatal
                End Try
            Next
        End If

        ' Let's keep the memory footprint down... 
        If dt.Rows.Count > My.Settings.RowBuffer Then
            fm.lbStatus.Text &= " (saving data)"
            Application.DoEvents()
            db.UpdateTable(dt)
            row_count += dt.Rows.Count
            dt.Clear()
        End If

        ' Check directories for errors
        Try
            di_array = p_di.GetDirectories
        Catch ex As Exception
            ' This probably redundant, since we catch this type of
            ' error above while processing files.
            dr = dt.NewRow
            dr("Server") = server
            dr("Path") = FormatPath(StartingDir)
            dr("Name") = Left("Error: " & ex.Message, 128)
            dt.Rows.Add(dr)
            Exit Sub
        End Try

        ' Process directories (and descend into the dir structure)
        For Each di In di_array
            Try
                ds = di.GetAccessControl()
                fsar = ds.GetAccessRules(True, True, GetType(SecurityIdentifier))

                ' let's see if this DACL is different from the DACL of the current directory
                isDifferent = False
                If consolidate Then
                    If fsar.Count <> LastDACL.Count Then
                        isDifferent = True
                    End If

                    ' dang, OK... we'll have to check each entry
                    If Not isDifferent Then
                        i = 0
                        For Each ar As FileSystemAccessRule In fsar
                            If ar.Equals(LastDACL(i)) Then
                                isDifferent = True
                                Exit For
                            End If
                            i += 1
                        Next
                    End If
                End If

                If Not consolidate Or isDifferent Then
                    For Each ar As FileSystemAccessRule In fsar
                        dr = dt.NewRow
                        dr("Server") = server
                        dr("Path") = FormatPath(di.FullName)
                        dr("Type") = ar.AccessControlType.ToString
                        dr("Name") = TranslateSID.TranslateSidToName(server, CType(ar.IdentityReference, SecurityIdentifier))
                        dr("Permissions") = DirectoryMaskToString(ar.FileSystemRights)
                        dr("Inherited") = ar.IsInherited
                        dr("Scope") = ACEFlagToString(ar.InheritanceFlags, ar.PropagationFlags, ar.IsInherited)
                        dt.Rows.Add(dr)
                    Next
                End If
                d_count += 1
            Catch ex As Exception
                dr = dt.NewRow
                dr("Server") = server
                dr("Path") = FormatPath(di.FullName)
                dr("Name") = Left("Error: " & ex.Message, 128)
                dt.Rows.Add(dr)
                ' not fatal
            End Try

            ' recursive call to this subroutine
            Doit(di.FullName)
        Next
    End Sub

    '
    ' Format the Path field.  Converts a UNC path back to it's drive letter
    ' equivalent.
    '
    Private Function FormatPath(ByVal Path As String) As String
        If needs_conversion Then
            Static index As Integer = -1
            Static DriveLetter As String

            If index < 0 Then
                ' these values don't change, so we can make 'em static
                index = Path.IndexOf("$")
                DriveLetter = Path.Substring(index - 1, 1).ToUpper & ":"
            End If

            Return Left(DriveLetter & Path.Substring(index + 1), 4000)
        Else
            Return Left(Path, 4000)
        End If
    End Function

    '
    ' User friendly version of the access masks designed to read just like
    ' the permissions "tab" when looking at the properties of a file inside 
    ' the Windows Explorer
    '
    Function FileMaskToString(ByVal mask As FileSystemRights) As String
        Dim sb As New System.Text.StringBuilder

        Select Case mask
            Case FileSystemRights.FullControl
                Return ("Full Control")
            Case FileSystemRights.Modify
                Return ("Modify")
            Case FileSystemRights.ReadAndExecute
                Return ("Read & Execute")
            Case FileSystemRights.ExecuteFile
                Return ("Execute")
            Case FileSystemRights.Read
                Return ("Read")
            Case FileSystemRights.Write
                Return ("Write")
                ' generic permissions
            Case CType(AccessTypes.GENERIC_ALL, FileSystemRights)
                Return ("Generic Full Control")
            Case CType(AccessTypes.GENERIC_READ Or AccessTypes.GENERIC_WRITE Or AccessTypes.GENERIC_EXECUTE Or AccessTypes.DELETE, FileSystemRights)
                Return ("Generic Modify")
            Case CType(AccessTypes.GENERIC_READ Or AccessTypes.GENERIC_EXECUTE, FileSystemRights)
                Return ("Generic Read & Execute")
            Case CType(AccessTypes.GENERIC_EXECUTE, FileSystemRights)
                Return ("Generic Execute")
            Case CType(AccessTypes.GENERIC_READ, FileSystemRights)
                Return ("Generic Read")
            Case CType(AccessTypes.GENERIC_WRITE, FileSystemRights)
                Return ("Generic Write")
            Case Else
                ' ok... do it the hard way (in bit order)
                sb.Append("Special (0x")
                sb.Append(Hex(mask))
                sb.Append("): ")
                If CBool(mask And CType(AccessTypes.GENERIC_READ, FileSystemRights)) Then
                    sb.Append("Generic Read,")
                End If
                If CBool(mask And CType(AccessTypes.GENERIC_WRITE, FileSystemRights)) Then
                    sb.Append("Generic Write,")
                End If
                If CBool(mask And CType(AccessTypes.GENERIC_EXECUTE, FileSystemRights)) Then
                    sb.Append("Generic Execute,")
                End If
                If CBool(mask And CType(AccessTypes.GENERIC_ALL, FileSystemRights)) Then
                    sb.Append("Generic Full,")
                End If
                If CBool(mask And CType(AccessTypes.ACCESS_SYSTEM_SECURITY, FileSystemRights)) Then
                    sb.Append("System,")
                End If
                If CBool(mask And FileSystemRights.Synchronize) Then
                    sb.Append("Sync,")
                End If
                If CBool(mask And FileSystemRights.TakeOwnership) Then
                    sb.Append("Ownership,")
                End If
                If CBool(mask And FileSystemRights.ChangePermissions) Then
                    sb.Append("Change Perm,")
                End If
                If CBool(mask And FileSystemRights.ReadPermissions) Then
                    sb.Append("Read Perm,")
                End If
                If CBool(mask And FileSystemRights.Delete) Then
                    sb.Append("Delete,")
                End If
                If CBool(mask And FileSystemRights.WriteAttributes) Then
                    sb.Append("Write Attr,")
                End If
                If CBool(mask And FileSystemRights.ReadAttributes) Then
                    sb.Append("Read Attr,")
                End If
                If CBool(mask And FileSystemRights.ExecuteFile) Then
                    sb.Append("Execute,")
                End If
                If CBool(mask And FileSystemRights.WriteExtendedAttributes) Then
                    sb.Append("Write ExtAttr,")
                End If
                If CBool(mask And FileSystemRights.ReadExtendedAttributes) Then
                    sb.Append("Read ExtAttr,")
                End If
                If CBool(mask And FileSystemRights.AppendData) Then
                    sb.Append("Append,")
                End If
                If CBool(mask And FileSystemRights.WriteData) Then
                    sb.Append("Write,")
                End If
                If CBool(mask And FileSystemRights.ReadData) Then
                    sb.Append("Read,")
                End If
                Return Left(sb.ToString.TrimEnd(","c), 255)
        End Select
    End Function

    '
    ' Same as above, but with slightly different wording
    '
    Function DirectoryMaskToString(ByVal mask As FileSystemRights) As String
        Dim sb As New System.Text.StringBuilder

        Select Case mask
            Case FileSystemRights.FullControl
                Return ("Full Control")
            Case FileSystemRights.Modify
                Return ("Modify")
            Case FileSystemRights.ReadAndExecute
                Return ("Read & Execute")
            Case FileSystemRights.ListDirectory
                Return ("List Folder Contents")
            Case FileSystemRights.Read
                Return ("Read")
            Case FileSystemRights.Write
                Return ("Write")
                ' generic permissions
            Case CType(AccessTypes.GENERIC_ALL, FileSystemRights)
                Return ("Generic Full Control")
            Case CType(AccessTypes.GENERIC_READ Or AccessTypes.GENERIC_WRITE Or AccessTypes.GENERIC_EXECUTE Or AccessTypes.DELETE, FileSystemRights)
                Return ("Generic Modify")
            Case CType(AccessTypes.GENERIC_READ Or AccessTypes.GENERIC_EXECUTE, FileSystemRights)
                Return ("Generic Read & Execute")
            Case CType(AccessTypes.GENERIC_EXECUTE, FileSystemRights)
                Return ("Generic List Folder Contents")
            Case CType(AccessTypes.GENERIC_READ, FileSystemRights)
                Return ("Generic Read")
            Case CType(AccessTypes.GENERIC_WRITE, FileSystemRights)
                Return ("Generic Write")
            Case Else
                ' ok... do it the hard way (in bit order)
                sb.Append("Special (0x")
                sb.Append(Hex(mask))
                sb.Append("): ")
                If CBool(mask And CType(AccessTypes.GENERIC_READ, FileSystemRights)) Then
                    sb.Append("Generic Read,")
                End If
                If CBool(mask And CType(AccessTypes.GENERIC_WRITE, FileSystemRights)) Then
                    sb.Append("Generic Write,")
                End If
                If CBool(mask And CType(AccessTypes.GENERIC_EXECUTE, FileSystemRights)) Then
                    sb.Append("Generic List,")
                End If
                If CBool(mask And CType(AccessTypes.GENERIC_ALL, FileSystemRights)) Then
                    sb.Append("Generic Full,")
                End If
                If CBool(mask And CType(AccessTypes.ACCESS_SYSTEM_SECURITY, FileSystemRights)) Then
                    sb.Append("System,")
                End If
                If CBool(mask And FileSystemRights.Synchronize) Then
                    sb.Append("Sync,")
                End If
                If CBool(mask And FileSystemRights.TakeOwnership) Then
                    sb.Append("Ownership,")
                End If
                If CBool(mask And FileSystemRights.ChangePermissions) Then
                    sb.Append("Change Perm,")
                End If
                If CBool(mask And FileSystemRights.ReadPermissions) Then
                    sb.Append("Read Perm,")
                End If
                If CBool(mask And FileSystemRights.Delete) Then
                    sb.Append("Delete,")
                End If
                If CBool(mask And FileSystemRights.WriteAttributes) Then
                    sb.Append("Write Attr,")
                End If
                If CBool(mask And FileSystemRights.ReadAttributes) Then
                    sb.Append("Read Attr,")
                End If
                If CBool(mask And FileSystemRights.DeleteSubdirectoriesAndFiles) Then
                    sb.Append("Delete Folders,")
                End If
                If CBool(mask And FileSystemRights.Traverse) Then
                    sb.Append("Traverse,")
                End If
                If CBool(mask And FileSystemRights.WriteExtendedAttributes) Then
                    sb.Append("Write ExtAttr,")
                End If
                If CBool(mask And FileSystemRights.ReadExtendedAttributes) Then
                    sb.Append("Read ExtAttr,")
                End If
                If CBool(mask And FileSystemRights.CreateDirectories) Then
                    sb.Append("Create Folders,")
                End If
                If CBool(mask And FileSystemRights.CreateFiles) Then
                    sb.Append("Create Files,")
                End If
                If CBool(mask And FileSystemRights.ListDirectory) Then
                    sb.Append("List,")
                End If
                Return Left(sb.ToString.TrimEnd(","c), 255)
        End Select

    End Function

    '
    ' A User Friendly version of the "Apply To" column
    '
    Function ACEFlagToString(ByVal inherit As InheritanceFlags, ByVal propagate As PropagationFlags, ByVal isInherited As Boolean) As String
        Dim flag As AceFlags
        Dim sb As New System.Text.StringBuilder

        ' let's reconstitue the ACE_Flag (without the INHERITED_ACE)
        flag = 0
        If CBool(inherit And InheritanceFlags.ContainerInherit) Then
            flag = flag Or AceFlags.CONTAINER_INHERIT_ACE
        End If
        If CBool(inherit And InheritanceFlags.ObjectInherit) Then
            flag = flag Or AceFlags.OBJECT_INHERIT_ACE
        End If
        If CBool(propagate And PropagationFlags.InheritOnly) Then
            flag = flag Or AceFlags.INHERIT_ONLY_ACE
        End If
        If CBool(propagate And PropagationFlags.NoPropagateInherit) Then
            flag = flag Or AceFlags.NO_PROPAGATE_INHERIT_ACE
        End If

        Select Case flag
            Case 0
                Return "This folder only"
            Case AceFlags.CONTAINER_INHERIT_ACE Or AceFlags.OBJECT_INHERIT_ACE
                Return "This folder, subfolders and files"
            Case AceFlags.CONTAINER_INHERIT_ACE
                Return "This folder and subfolders"
            Case AceFlags.OBJECT_INHERIT_ACE
                Return "This folder and files"
            Case AceFlags.CONTAINER_INHERIT_ACE Or AceFlags.OBJECT_INHERIT_ACE Or AceFlags.INHERIT_ONLY_ACE
                Return "Subfolders and files only"
            Case AceFlags.CONTAINER_INHERIT_ACE Or AceFlags.INHERIT_ONLY_ACE
                Return "Subfolders only"
            Case AceFlags.OBJECT_INHERIT_ACE Or AceFlags.INHERIT_ONLY_ACE
                Return "Files only"
            Case Else
                ' ok... do it the hard way (in bit order)
                ' Add INHERITED_ACE back in so the bit mask is correct
                If isInherited Then
                    flag = flag Or AceFlags.INHERITED_ACE
                End If
                sb.Append("Special (0x")
                sb.Append(Hex(flag))
                sb.Append("): ")
                If CBool(flag And AceFlags.INHERITED_ACE) Then
                    sb.Append("Inherited")
                End If
                If CBool(flag And AceFlags.INHERIT_ONLY_ACE) Then
                    sb.Append("Inherit Only,")
                End If
                If CBool(flag And AceFlags.INHERIT_ONLY_ACE) Then
                    ' there are already too many people in the world :)
                    sb.Append("No Propagate,")
                End If
                If CBool(flag And AceFlags.CONTAINER_INHERIT_ACE) Then
                    sb.Append("Container,")
                End If
                If CBool(flag And AceFlags.OBJECT_INHERIT_ACE) Then
                    sb.Append("Object,")
                End If
                Return Left(sb.ToString.TrimEnd(","c), 128)
        End Select
    End Function
End Class
