Imports System.IO

Public Class fmMain
    Private Sub fmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        ' just fill in the blanks with default values
        tbServer.Text = Environment.MachineName
        tbTableName.Text = Format(Now(), "yyMMdd")
    End Sub

    Private Sub bnBrowseDir_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnBrowseDir.Click
        FolderBrowserDialog1.ShowNewFolderButton = False
        If FolderBrowserDialog1.ShowDialog() = DialogResult.OK Then
            tbStartDir.Text = FolderBrowserDialog1.SelectedPath
        End If
    End Sub

    Private Sub bnBrowseDb_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnBrowseDb.Click
        SaveFileDialog1.AddExtension = True
        SaveFileDialog1.FileName = tbDatabase.Text
        SaveFileDialog1.DefaultExt = "MDB"
        SaveFileDialog1.Filter = "Microsoft Access files (*.mdb)|*.mdb|All files (*.*)|*.*"
        SaveFileDialog1.InitialDirectory = tbDatabase.Text
        SaveFileDialog1.CheckFileExists = False
        SaveFileDialog1.OverwritePrompt = False
        If SaveFileDialog1.ShowDialog() = DialogResult.OK Then
            tbDatabase.Text = SaveFileDialog1.FileName
        End If
    End Sub

    Private Sub bnQuit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnQuit.Click
        Me.Close()
    End Sub

    Private Sub bnDoit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles bnDoit.Click
        Dim db As AccessDatabase
        Dim dt As DataTable
        Dim total_rows As Integer = 0
        Dim DDL As String

        ' quick sanity check for user input
        If tbServer.Text = "" Or tbDatabase.Text = "" Or tbstartdir.Text = "" Or tbTableName.Text = "" Then
            MsgBox("You must provide input in all 4 boxes", MsgBoxStyle.Critical, "Error")
            Exit Sub
        End If

        ' another sanity check...
        If DirectoryEx.Exists(ConvToUNC(tbServer.Text, tbStartDir.Text)) = False Then
            MsgBox("Can not access the Starting Directory", MsgBoxStyle.Critical, "Error")
            Exit Sub
        End If

        ' Even more sanity checking... does the server name match the server in the UNC path?
        If tbStartDir.Text.StartsWith("\\") Then
            Dim parts() As String
            parts = tbStartDir.Text.Split("\"c)
            If parts.Length > 3 Then
                If String.Compare(parts(2), tbServer.Text, True) <> 0 Then
                    MsgBox("Server Name doesn't match the UNC Path for Starting Directory", MsgBoxStyle.Critical, "Error")
                    Exit Sub
                End If
            End If
        End If

        db = New AccessDatabase
        Try
            ' Does the database already exist?  If not, create it!
            db.CreateDatabase(tbDatabase.Text)
        Catch ex As Exception
            MsgBox("Can not create the database" & vbCr & ex.Message, MsgBoxStyle.Critical, "Error")
            Exit Sub
        End Try

        ' Microsoft Access is limited to 255 character for the Text type, but the Path
        ' field needs to be a whole lot bigger (I use a limit of 4000 inside the traverse
        ' class to be compatible with SQL Server's varchar field).  The problem is that
        ' Access can't sort or group on anything but the first 255 characters of a memo field.
        DDL = "(ID int identity primary key, " _
        & "Server varchar(50), " _
        & "Path memo, " _
        & "Type varchar(10), " _
        & "Name varchar(128), " _
        & "Permissions varchar(255), " _
        & "Inherited bit, " _
        & "Scope varchar(128))"
        Try
            If db.TableExists(tbDatabase.Text, tbTableName.Text) Then
                Select Case MsgBox("Database table already exists, Overwrite the contents?", MsgBoxStyle.YesNoCancel, "Warning")
                    Case MsgBoxResult.Yes
                        db.EmptyTable(tbDatabase.Text, tbTableName.Text)
                    Case MsgBoxResult.No
                        ' do nothing
                    Case MsgBoxResult.Cancel
                        Exit Sub
                End Select
            End If

            ' Create a new table (if we need to)
            db.CreateTable(tbDatabase.Text, tbTableName.Text, DDL)
        Catch ex As Exception
            MsgBox("Can not create the table" & vbCr & ex.Message, MsgBoxStyle.Critical, "Error")
            Exit Sub
        End Try

        ' Open the database
        Try
            dt = db.OpenTable(tbDatabase.Text, tbTableName.Text)
        Catch ex As Exception
            MsgBox("Can not open the database" & vbCr & ex.Message, MsgBoxStyle.Critical, "Error")
            Exit Sub
        End Try

        Dim tr As New Traverse
        tr.IsFirst = True
        tr.DirOnly = cbDirOnly.Checked
        tr.ConsolidateDirs = cbConsoldiate.Checked
        tr.MainForm = Me
        tr.AccessDatabase = db
        tr.DataTable = dt
        tr.ServerName = tbServer.Text
        tr.RecordRootName = tbStartDir.Text

        ' Do the deed
        Cursor.Current = System.Windows.Forms.Cursors.WaitCursor
        Try
            tr.Doit(ConvToUNC(tbServer.Text, tbStartDir.Text))
        Catch ex As Exception
            MsgBox("Internal Error" & vbCr & ex.Message, MsgBoxStyle.Critical, "Error")
            Exit Sub
        End Try

        ' save any pending "end of buffer" data
        Try
            If dt.Rows.Count > 0 Then
                lbStatus.Text &= " (saving data)"
                Application.DoEvents()
                db.UpdateTable(dt)
            End If
        Catch ex As Exception
            MsgBox("Can not update the database" & vbCr & ex.Message, MsgBoxStyle.Critical, "Error")
            Exit Sub
        End Try

        Cursor.Current = System.Windows.Forms.Cursors.Default
        lbStatus.Text = "Done"
        MsgBox("Done!" & vbCr & "Processed " & tr.DirectoryCount & " directories and " & tr.FileCount & " files" & vbCr & _
               "Recorded " & tr.RowCount + dt.Rows.Count & " entries into the database", MsgBoxStyle.Information, "Success")
    End Sub

    '
    ' Convert a Remote PC name, Local path name pair into a UNC style path
    '
    Private Function ConvToUNC(ByVal RemotePC As String, ByVal Path As String) As String

        ' no need for UNC if we're on the same PC
        If RemotePC = "" Or String.Compare(RemotePC, Environment.MachineName, True) = 0 Then
            Return Path
        End If

        ' was the name provided already a UNC path?
        If Path.StartsWith("\\") Then
            Return Path
        End If

        If Path.Length > 2 Then
            Dim sb As New System.Text.StringBuilder

            sb.Append("\\")
            sb.Append(RemotePC)
            sb.Append("\")
            sb.Append(Path.Substring(0, 1))
            sb.Append("$")
            sb.Append(Path.Substring(2))
            Return sb.ToString
        End If

        Return Path
    End Function

    Private Sub cbDirOnly_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cbDirOnly.CheckedChanged
        If Not cbDirOnly.Checked And cbConsoldiate.Checked Then
            cbConsoldiate.Checked = False
        End If
    End Sub

    Private Sub cbConsoldiate_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cbConsoldiate.CheckedChanged
        If cbConsoldiate.Checked And Not cbDirOnly.Checked Then
            cbDirOnly.Checked = True
        End If
    End Sub
End Class